﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace WindowsFormsApplication1
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        double _eQuatorialEarthRadius = 6371D;
        double _d2r = (Math.PI / 180D);

        private int HaversineInM(double lat1, double long1, double lat2, double long2)
        {
            return (int)(1000D * HaversineInKM(lat1, long1, lat2, long2));
        }

        private double HaversineInKM(double lat1, double long1, double lat2, double long2)
        {
            double dlong = (long2 - long1) * _d2r;
            double dlat = (lat2 - lat1) * _d2r;
            double a = Math.Pow(Math.Sin(dlat / 2D), 2D) + Math.Cos(lat1 * _d2r) * Math.Cos(lat2 * _d2r) * Math.Pow(Math.Sin(dlong / 2D), 2D);
            double c = 2D * Math.Atan2(Math.Sqrt(a), Math.Sqrt(1D - a));
            double d = _eQuatorialEarthRadius * c;

            return d;
        }

        private struct common_header
        {
            public short accelx;
            public short accely;
            public short accelz;
            public byte type;
            public int result;

        }

        private struct one_header
        {
            public byte[] motor;
            public bool run_motor;
            public bool fail_save;
            public byte gps_sat_common;
            public byte flight_mode;
            public byte gps_sat_not_use;
            public byte gps_status;
            public bool init_ok;
            public int result;

        }

        private struct two_header
        {

            public Int32 home_lon;
            public Int32 home_lat;
            public short basic_altitude;
            public Int32 fly_lon;
            public Int32 fly_lat;
            public short fly_altitude;
            public short vert_speed;

            public short x_gorizontspeed;
            public short y_gorizontspeed;

            public short pitch;
            public short pitch_torque;
            public short roll;
            public short roll_torque;
            public short yaw_torque;

            public int result;
        }

        private struct three_header
        {
            public short altitude_baro1;
            public short altitude_baro2;
            public ushort voltage;
            public short gyrox;
            public short gyroy;
            public short gyroz;
            public int result;
        }

        private struct four_header
        {
            public UInt64 time_milli;
            public int result;
        }

        private four_header read_four_header(FileStream fs)
        {
            four_header fh = new four_header() { result = 1 };
            byte[] buffer = new byte[40];
            if (fs.Read(buffer, 0, 40) < 40) return fh;

            fh.time_milli = BitConverter.ToUInt64(buffer, 0);
            fh.result = 0;
            return fh;
        }

        private three_header read_three_header(FileStream fs)
        {
            three_header th = new three_header() { result = 1 };
            byte[] buffer = new byte[40];
            if (fs.Read(buffer, 0, 40) < 40) return th;

            th.altitude_baro1= BitConverter.ToInt16(buffer, 18);
            th.voltage = BitConverter.ToUInt16(buffer, 22);
            th.altitude_baro2 = BitConverter.ToInt16(buffer, 24);
            th.gyrox=BitConverter.ToInt16(buffer, 32);
            th.gyroy = BitConverter.ToInt16(buffer, 34);
            th.gyroz = BitConverter.ToInt16(buffer, 36);
            th.result = 0;
            return th;
        }


        private two_header read_two_header(FileStream fs)
        {
            two_header th = new two_header() { result = 1 };
            byte[] buffer = new byte[40];
            if (fs.Read(buffer, 0, 40) < 40) return th;
            th.home_lon = BitConverter.ToInt32(buffer, 0);
            th.home_lat = BitConverter.ToInt32(buffer, 4);
            th.basic_altitude = BitConverter.ToInt16(buffer, 8);
            th.fly_lon= BitConverter.ToInt32(buffer, 10);
            th.fly_lat = BitConverter.ToInt32(buffer, 14);
            th.fly_altitude = BitConverter.ToInt16(buffer, 18);
            th.vert_speed= BitConverter.ToInt16(buffer, 24);
            th.roll = BitConverter.ToInt16(buffer, 28);
            th.roll_torque = BitConverter.ToInt16(buffer, 26);
            th.pitch = BitConverter.ToInt16(buffer, 32);
            th.pitch_torque = BitConverter.ToInt16(buffer, 30);
            th.yaw_torque = BitConverter.ToInt16(buffer, 34);

            th.x_gorizontspeed = BitConverter.ToInt16(buffer, 20);
            th.y_gorizontspeed = BitConverter.ToInt16(buffer, 22);

            th.result = 0;
            return th;
        }

        private one_header read_one_header(FileStream fs)
        {
            one_header oh=new one_header() { result = 1 };
            byte[] buffer = new byte[40];
            if (fs.Read(buffer, 0, 40) < 40) return oh;
            oh.motor = new byte[8];
            oh.motor[0] = buffer[15];
            oh.motor[1] = buffer[16];
            oh.motor[2] = buffer[17];
            oh.motor[3] = buffer[18];
            oh.motor[4] = buffer[19];
            oh.motor[5] = buffer[20];
            oh.motor[6] = buffer[21];
            oh.motor[7] = buffer[22];
            if (buffer[27] == 0) oh.run_motor = false; else oh.run_motor = true;
            if (buffer[28] == 0) oh.fail_save = false; else oh.fail_save = true;
            oh.gps_sat_common = buffer[33];
            oh.flight_mode= buffer[30];
            oh.gps_status= buffer[34];
            oh.gps_sat_not_use= buffer[35];
            if (buffer[36] == 0) oh.init_ok = false; else oh.init_ok = true;

            oh.result = 0;
            return oh;

        }

        private common_header read_common_header(FileStream fs)
        {
            common_header ch = new common_header() { result = 1};
            byte[] buffer=new byte[24];
            if (fs.Read(buffer, 0, 24)<24) return ch;
            if (BitConverter.ToUInt16(buffer,0)!=(ushort)0x6bd2) return ch;
            ch.accelx = BitConverter.ToInt16(buffer, 17);
            ch.accely = BitConverter.ToInt16(buffer, 19);
            ch.accelz = BitConverter.ToInt16(buffer, 21);
            ch.type = buffer[23];
            ch.result = 0;
            return ch;
        }

        private void button2_Click(object sender, EventArgs e)
        {
            openFileDialog1.ShowDialog();
        }

        private void openFileDialog1_FileOk(object sender, CancelEventArgs e)
        {
            textBox1.Text = openFileDialog1.FileName;
            textBox2.Text = Path.ChangeExtension(openFileDialog1.FileName,".csv");
        }

        private void button1_Click(object sender, EventArgs e)
        {
            FileStream fs;
            double voltage_offset;

            try
            {
                voltage_offset = double.Parse(textBox3.Text);
            }
            catch
            {
                MessageBox.Show("Bad voltage offset: " + textBox3.Text, "Error");
                return;
            }

            try {
                fs = File.Open(textBox1.Text, FileMode.Open, FileAccess.Read);
            }
            catch
            {
                MessageBox.Show("Error open input file: " + textBox1.Text, "Error");
                return;
            }

            common_header ch1,ch2, ch3, ch4;
            one_header one_head;
            two_header two_head;
            three_header three_head;
            four_header four_head;

            UInt64 base_time=0;

            do
            {
                ch1 = read_common_header(fs);
                one_head = new one_header() { init_ok = false };
                switch (ch1.type)
                {
                    case 1:
                        one_head = read_one_header(fs);
                        if (one_head.result != 0)
                        {
                            MessageBox.Show("Initialization not find: " + textBox1.Text, "Error");
                            fs.Close();
                            return;
                        }

                        break;
                    case 2:
                        fs.Seek(40, SeekOrigin.Current);
                        break;
                    case 3:
                        fs.Seek(40, SeekOrigin.Current);
                        break;
                    case 4:
                        fs.Seek(40, SeekOrigin.Current);
                        break;
                    default:
                        MessageBox.Show("Bad format: " + textBox1.Text, "Error");
                        fs.Close();
                        return;
                }
            } while (!one_head.init_ok);

            FileStream outfs;
            try
            {
                outfs = File.Open(textBox2.Text, FileMode.Create, FileAccess.Write);
            }
            catch
            {
                MessageBox.Show("Error open output file: " + textBox2.Text, "Error");
                fs.Close();
                return;
            }

            //запись лога
            StreamWriter writer = new StreamWriter(outfs);

            int motor_count=0;
            foreach(byte i in one_head.motor)
            {
                if (i != 0x69) motor_count++; else break;
            }


            //запись шапки
            writer.Write("time(millis)");
            writer.Write(';');
            writer.Write("mode");
            writer.Write(';');
            writer.Write("altitude(m)");
            writer.Write(';');
            writer.Write("VSpeed(m/s)");
            writer.Write(';');
            writer.Write("GSpeed(m/s)");
            writer.Write(';');
            writer.Write("distance(m)");
            writer.Write(';');
            writer.Write("fly_gps");
            writer.Write(';');
            writer.Write("home_gps");
            writer.Write(';');
            writer.Write("satellites_used");
            writer.Write(';');
            writer.Write("voltage");
            writer.Write(';');
            writer.Write("F/S");
            writer.Write(';');
            writer.Write("pitch");
            writer.Write(';');
            writer.Write("roll");
            writer.Write(';');
            writer.Write("pitch_torque");
            writer.Write(';');
            writer.Write("roll_torque");            
            writer.Write(';');
            writer.Write("yaw_torque");
            writer.Write(';');
            writer.Write("gyroscope");
            writer.Write(';');
            writer.Write("acceleration");
            writer.Write(';');
            for (int i=0;i< motor_count; i++)
            {
                writer.Write("motor_"+(i+1).ToString());
                writer.Write(';');
            }



            writer.WriteLine();
            writer.Flush();

            bool is_first_block = true;

            while (true)
            {
                if (!is_first_block)
                {
                    ch1 = read_common_header(fs);
                    if ((ch1.result != 0)||(ch1.type!=1)) break;
                    one_head = read_one_header(fs);
                    if (one_head.result != 0) break;
                }
                
                ch2 = read_common_header(fs);
                if ((ch2.result != 0)|| (ch2.type != 2)) break;
                two_head = read_two_header(fs);
                if (two_head.result != 0) break;

                ch3 = read_common_header(fs);
                if ((ch3.result != 0)|| (ch3.type != 3)) break;
                three_head = read_three_header(fs);
                if (three_head.result != 0) break;

                ch4 = read_common_header(fs);
                if ((ch4.result != 0)|| (ch4.type != 4)) break;
                four_head = read_four_header(fs);
                if (four_head.result != 0) break;

                if (is_first_block)
                {
                    base_time = four_head.time_milli;
                    is_first_block = false;
                }
                    

                //время
                writer.Write(four_head.time_milli- base_time);
                writer.Write(';');

                //полетный режим
                switch (one_head.flight_mode)
                {
                    case 1: writer.Write("MAN");
                        break;
                    case 2: writer.Write("ATTI");
                        break;
                    case 3: writer.Write("GPS");
                        break;
                    default: writer.Write("N/A");
                        break;
                }
                writer.Write(';');

                //высота
                double altitude = (two_head.fly_altitude - two_head.basic_altitude)/10D;
                writer.Write("=\""+altitude.ToString("N1")+'"');
                writer.Write(';');

                //вертикальная скорость
                writer.Write("=\"" + (two_head.vert_speed / 10D).ToString("N1") + '"');
                writer.Write(';');

                //горизонтальная скорость
                writer.Write("=\"" + (Math.Sqrt(Math.Pow(two_head.x_gorizontspeed, 2D) + Math.Pow(two_head.y_gorizontspeed, 2D))/10D).ToString("N2") + '"');
                writer.Write(';');

                



                //дальность
                double fly_lat = two_head.fly_lat / 10000000D;
                double fly_lon = two_head.fly_lon / 10000000D;
                double home_lat = two_head.home_lat / 10000000D;
                double home_lon = two_head.home_lon / 10000000D;

                
                if (one_head.gps_status < 2)
                {
                    writer.Write("N/A");
                }
                else {
                    writer.Write("=\"" + HaversineInM(fly_lat, fly_lon, home_lat, home_lon).ToString() + '"');
                }
                writer.Write(';');

                //координаты ЛА
                if (one_head.gps_status<1){
                    writer.Write("N/A");
                }else{ 
                 writer.Write("=\"" + fly_lat.ToString("N7", System.Globalization.CultureInfo.GetCultureInfo("en-US")));
                 writer.Write(',');
                 writer.Write(fly_lon.ToString("N7", System.Globalization.CultureInfo.GetCultureInfo("en-US"))+'"');
                }
                writer.Write(';');

                //координаты home
                if (one_head.gps_status < 2)
                {
                   writer.Write("N/A");
                }
                else {
                    writer.Write("=\"" + home_lat.ToString("N7", System.Globalization.CultureInfo.GetCultureInfo("en-US")));
                    writer.Write(',');
                    writer.Write(home_lon.ToString("N7", System.Globalization.CultureInfo.GetCultureInfo("en-US"))+'"');
                }
                writer.Write(';');

                //количество используемых спутников
                writer.Write(one_head.gps_sat_common - one_head.gps_sat_not_use);
                writer.Write(';');
                //напряжение
                writer.Write("=\"" + (three_head.voltage / 100D + voltage_offset).ToString("N2")+'"');
                writer.Write(';');
                //fail save
                if (one_head.fail_save) writer.Write("yes"); else writer.Write("no");
                writer.Write(';');

                //pitch
                writer.Write("=\"" + (two_head.pitch / 10D).ToString("N1") + '"');
                writer.Write(';');

                //roll
                writer.Write("=\"" + (two_head.roll / 10D).ToString("N1") + '"');
                writer.Write(';');

                //pitch_torque
                writer.Write("=\"" + (two_head.pitch_torque / 10D).ToString("N1") + '"');
                writer.Write(';');

                //roll_torque
                writer.Write("=\"" + (two_head.roll_torque / 10D).ToString("N1") + '"');
                writer.Write(';');

                //yaw_torgue
                writer.Write("=\"" + (two_head.yaw_torque / 10D).ToString("N1") + '"');
                writer.Write(';');

                //gyroscope
                writer.Write('"' + three_head.gyrox.ToString() +';'+ three_head.gyroy.ToString() +';'+ three_head.gyroz.ToString()+'"');
                writer.Write(';');
                //acceleration
                int accelx = (ch1.accelx + ch2.accelx + ch3.accelx + ch4.accelx) / 4;
                int accely = (ch1.accely + ch2.accely + ch3.accely + ch4.accely) / 4;
                int accelz = (ch1.accelz + ch2.accelz + ch3.accelz + ch4.accelz) / 4;
                writer.Write('"' + accelx.ToString() + ';' + accely.ToString() + ';' + accelz.ToString()+'"');

                
                //моторы


                if (one_head.run_motor)
                {
                    for (int i = 0; i < motor_count; i++)
                    {
                        writer.Write(';');
                        writer.Write(one_head.motor[i] - 0x69);

                    }
                }
                else
                {
                    for (int i = 0; i < motor_count; i++)
                    {
                        writer.Write(';');
                        writer.Write('0');

                    }

                }

                writer.WriteLine();
                writer.Flush();

            }

            writer.Close();
            fs.Close();


            DialogResult dr=MessageBox.Show("Saved successfully: "+textBox2.Text+"\nOpen?", "OK",MessageBoxButtons.OKCancel);

            if (dr == DialogResult.OK)
            {
                Process.Start(textBox2.Text);
            }



            


        }

        private void textBox1_DragDrop(object sender, DragEventArgs e)
        {
            string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);
            if (files != null && files.Length != 0)
            {
                textBox1.Text = files[0];
                textBox2.Text = Path.ChangeExtension(files[0], ".csv");
            }
        }

        private void textBox1_DragOver(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
                e.Effect = DragDropEffects.Copy;
            else
                e.Effect = DragDropEffects.None;
        }
    }
}
